#include <stdio.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <signal.h>
#include <X11/Xlib.h>

#include "syre.h"

#define FPS 10
#define FRAMETIME 1000000/FPS

#define NUM_OVERTONES 16

/*#define _USE_SIGNAL*/

Display *disp;
Window Root; 
Window win;
Visual *vis; 
int scr; 
unsigned int depth;
GC fg,bg;
int blk,wht;
int width,height;

extern double note_rates[96];

extern int dev_dsp;
extern bit16 sinwave[1024];

bit16 playbuffer16[2*PLAYBACK_FREQ]; /* one second of sound... */
bit32 waveform[1024];

int sequence[16];

int playlength=2*PLAYBACK_FREQ; /* length of playbuffer */
int vol=48 /* 0-64 */, pan=16, lpan=16, rpan=16; /* 0-32 */
int attack=16, release=48; /* 0-64 , attack<=release */
int portamento=0; /* 0-64 */
int bpm4=142*4; /* 4 * beats per minute */
int seqi=0; /* sequence index */
int ci=0; /* chaos index :) */
int beat; /* number of sample values per beat */
int playing=0; /* boolean */
int last_space=0; /* boolean (for on/off toggle) */

typedef struct {
   int si,r;
   int vol;
} harmonic;

harmonic overtones[NUM_OVERTONES];

#ifdef _USE_SIGNAL
void handler_alarm(int sig)
#else
void tick(void)
#endif
{   
   struct itimerval itv;
   XEvent report;
   Window proot, pchild;
   int rootx, rooty, winx, winy;
   unsigned int keys_buttons;
   int dx, dy;
   char keys[32];
   unsigned int i, j, k;
   int rate;
   bit32 di[2];
   bit32 d;
   int wi;
   int attack_sub, release_sub; /* subscripts (indices) of
			       * attack end, release start
			       */
   int attack_delta, release_delta;
   int envelope=0;
   int port_sub, port_delta, port_val;
   
#ifdef _USE_SIGNAL
   signal(sig, handler_alarm);
#endif
   
   XQueryKeymap(disp, keys);

/*
    for(i=0; i<32; i++) {
       for(j=0; j<8; j++) {
	  if(keys[i]&(1<<j)) {
	     fprintf(stderr, "Key %d pressed [%d,%d]\n", i*8 + j, i, j);
	  }
       }
    }
*/


#ifdef KEYNOTES
   if(keys[3]&1) {
      playing=1;
      rate=123;
   }
   /* Y ==> 'A' */
   else if(keys[3]&(1<<5)) {
      playing=1;
      rate=220;
   }
   
   else {
      playing=0;
   }
#endif

   /* Enter chaos */
   if(keys[4]&(1<<4)) /* ENTER */{
      chaos();
/*      chaos_harmonics();*/
   }
   if(keys[7]&(1<<6)) /* RIGHT SHIFT */{
      chaos_harmonics();
   }
   if(keys[6]&(1<<2)) /* LEFT SHIFT */{
      clear_harmonics();
   }
   
   /* ON/OFF */
   if(!last_space && (keys[8]&(1<<1))) {
      playing=1-playing;
      last_space=1;
   }
   else if(last_space && (!(keys[8]&(1<<1)))) {
      last_space=0;
   }
   
   
   /* BPM */
   if(keys[2]&(1<<4)) {
      bpm4--; if(bpm4<120) bpm4=120;
   }
   else if(keys[2]&(1<<5)) {
      bpm4++; if(bpm4>960) bpm4=960;
   }
   else if(keys[4]&(1<<2)) {
      bpm4-=4; if(bpm4<120) bpm4=120;
   }
   else if(keys[4]&(1<<3)) {
      bpm4+=4; if(bpm4>960) bpm4=960;
   }
   beat =60*PLAYBACK_FREQ/bpm4; /* samples per beat */

   
   /* VOL */
   if(keys[12]&(1<<2)) /* UP ARROW */{
      vol++; if(vol>64) vol=64;
   }
   else if(keys[13]&1) /* DOWN ARROW */{
      vol--; if(vol<0) vol=0;
   }

   /* PAN */
   if(keys[12]&(1<<4)) /* LEFT ARROW */{
      pan--; if(pan<0) pan=0;
      if(pan<=16) {
	 lpan=pan; rpan=16;
      }
      else {
	 lpan=16; rpan=31-pan; 
      }
      
   }
   else if(keys[12]&(1<<6)) /* RIGHT ARROW */{
      pan++; if(pan>31) pan=31;
      if(pan<=16) {
	 lpan=pan; rpan=16;
      }
      else {
	 lpan=16; rpan=31-pan;
      }
   }

   /* ENVELOPE */
   if(keys[7]*(1<<4)) /* .> */{
      attack--; if(attack<0) attack=0;
   }
   else if(keys[5]&(1<<6)) /* L */{
      attack++; if(attack>63) attack=63;
      if(attack>release) release=attack;
   }
   if(keys[7]*(1<<5)) /* / */{
      release++; if(release>63) release=63;
   }
   else if(keys[5]&(1<<7)) /* ; */{
      release--; if(release<0) release=0;
      if(release<attack) attack=release;
   }

   /* PORTAMENTO */
   if(keys[2]*(1<<3)) /* 0 */{
      portamento++; if(portamento>63) portamento=63;
   }
   else if(keys[4]&(1<<1)) /* P */{
      portamento--; if(portamento<0) portamento=0;
   }
   
   /* harmonics */
   if(keys[4]&(1<<6)) {
      overtones[0].vol++; if(overtones[0].vol>32) overtones[0].vol=32;
   }
   else if(keys[6]&(1<<4)) {
      overtones[0].vol--; if(overtones[0].vol<0) overtones[0].vol=0;
   }
   if(keys[4]&(1<<7)) {
      overtones[1].vol++; if(overtones[1].vol>32) overtones[1].vol=32;
   }
   else if(keys[6]&(1<<5)) {
      overtones[1].vol--; if(overtones[1].vol<0) overtones[1].vol=0;
   }
   if(keys[5]&(1)) {
      overtones[2].vol++; if(overtones[2].vol>32) overtones[2].vol=32;
   }
   else if(keys[6]&(1<<6)) {
      overtones[2].vol--; if(overtones[2].vol<0) overtones[2].vol=0;
   }
   if(keys[5]&(1<<1)) {
      overtones[3].vol++; if(overtones[3].vol>32) overtones[3].vol=32;
   }
   else if(keys[6]&(1<<7)) {
      overtones[3].vol--; if(overtones[3].vol<0) overtones[3].vol=0;
   }
   if(keys[5]&(1<<2)) {
      overtones[4].vol++; if(overtones[4].vol>32) overtones[4].vol=32;
   }
   else if(keys[7]&(1)) {
      overtones[4].vol--; if(overtones[4].vol<0) overtones[4].vol=0;
   }
   if(keys[5]&(1<<3)) {
      overtones[5].vol++; if(overtones[5].vol>32) overtones[5].vol=32;
   }
   else if(keys[7]&(1<<1)) {
      overtones[5].vol--; if(overtones[5].vol<0) overtones[5].vol=0;
   }
   if(keys[5]&(1<<4)) {
      overtones[6].vol++; if(overtones[6].vol>32) overtones[6].vol=32;
   }
   else if(keys[7]&(1<<2)) {
      overtones[6].vol--; if(overtones[6].vol<0) overtones[6].vol=0;
   }
   if(keys[5]&(1<<5)) {
      overtones[7].vol++; if(overtones[7].vol>32) overtones[7].vol=32;
   }
   else if(keys[7]&(1<<3)) {
      overtones[7].vol--; if(overtones[7].vol<0) overtones[7].vol=0;
   }

   if(keys[1]&(1<<2)) {
      overtones[8].vol++; if(overtones[8].vol>32) overtones[8].vol=32;
   }
   else if(keys[3]&(1)) {
      overtones[8].vol--; if(overtones[8].vol<0) overtones[8].vol=0;
   }
   if(keys[1]&(1<<3)) {
      overtones[9].vol++; if(overtones[9].vol>32) overtones[9].vol=32;
   }
   else if(keys[3]&(1<<1)) {
      overtones[9].vol--; if(overtones[9].vol<0) overtones[9].vol=0;
   }
   if(keys[1]&(1<<4)) {
      overtones[10].vol++; if(overtones[10].vol>32) overtones[10].vol=32;
   }
   else if(keys[3]&(1<<2)) {
      overtones[10].vol--; if(overtones[10].vol<0) overtones[10].vol=0;
   }
   if(keys[1]&(1<<5)) {
      overtones[11].vol++; if(overtones[11].vol>32) overtones[11].vol=32;
   }
   else if(keys[3]&(1<<3)) {
      overtones[11].vol--; if(overtones[11].vol<0) overtones[11].vol=0;
   }
   if(keys[1]&(1<<6)) {
      overtones[12].vol++; if(overtones[12].vol>32) overtones[12].vol=32;
   }
   else if(keys[3]&(1<<4)) {
      overtones[12].vol--; if(overtones[12].vol<0) overtones[12].vol=0;
   }
   if(keys[1]&(1<<3)) {
      overtones[13].vol++; if(overtones[13].vol>32) overtones[13].vol=32;
   }
   else if(keys[3]&(1<<5)) {
      overtones[13].vol--; if(overtones[13].vol<0) overtones[13].vol=0;
   }
   if(keys[2]&(1)) {
      overtones[14].vol++; if(overtones[14].vol>32) overtones[14].vol=32;
   }
   else if(keys[3]&(1<<6)) {
      overtones[14].vol--; if(overtones[14].vol<0) overtones[14].vol=0;
   }
   if(keys[2]&(1<<1)) {
      overtones[15].vol++; if(overtones[15].vol>32) overtones[15].vol=32;
   }
   else if(keys[3]&(1<<7)) {
      overtones[15].vol--; if(overtones[15].vol<0) overtones[15].vol=0;
   }

/*   overtones[0].r= sequence[seqi]*1024/(2*PLAYBACK_FREQ);*/

   /* ======> MOVE OUT OF LOOP!!!!!!! */
   overtones[0].r= 123*1024/(2*PLAYBACK_FREQ);
   for(i=1; i<16; i++) {
      overtones[i].r = overtones[0].r * (i+1);
   }

   /* CALCULATE WAVEFORM */
   for(i=0; i<1024; i++) {
      d=0;
      for(j=0; j<16; j++) {
	 d+= (sinwave[overtones[j].si] * overtones[j].vol)>>5;
/*	 d+= overtones[j].vol;*/
	 overtones[j].si += overtones[j].r;
	 overtones[j].si &= 1023;
      }
      waveform[i]=d;
   }
   
   /* CALCULATE ENVELOPE */
   attack_sub = (beat*2*attack)>>6;
   release_sub = (beat*2*release)>>6;
   if(attack_sub) attack_delta = 32768/attack_sub;
   if((beat<<1)-release_sub) release_delta = 32768/((beat<<1)-release_sub);

   if(attack) envelope=0;
   else envelope=32768;
   
   /* CALCULATE PORTAMENTO */
   port_sub = ((beat)*2*portamento)>>6;
   if(port_sub) port_delta = 32768/(port_sub);

   if(portamento) port_val=0;
   else port_val=32768;
   
   if(playing) {
      wi=0;
      /* ATTACK */
/*      for(i=0; i<attack_sub; i+=2) {*/
      for(i=0; i<beat*2; i+=2) {
	 playbuffer16[i] = (bit16)((waveform[wi>>6]*lpan*vol)>>11);
/*	 playbuffer16[i+1] = (bit16)((waveform[wi++]*rpan*vol)>>11);*/
	 playbuffer16[i+1] = (bit16)((waveform[wi>>6]*rpan*vol)>>11);
/*	 wi += sequence[seqi];*/
	 if(i<port_sub) {
	    wi +=
	      (port_val*sequence[seqi] +
	       (32768-port_val)*sequence[(seqi+15)%16])>>15;
	    port_val += port_delta;
	 }
	 else
	   wi += sequence[seqi];
	 
	 wi &= (1<<16)-1;
	 
#ifdef RECALC_WAVE
	 if(!(wi%=1024)) {
	    d=0;
	    for(j=0; j<16; j++) {
	       d+= (sinwave[overtones[j].si] * overtones[j].vol)>>5;
	       if(i<port_sub) {
		  overtones[j].si +=
		    (port_val*overtones[j].r +
		     (32768-port_val)*overtones[(j-1)%16].r)>>15;
		  port_val += port_delta;
	       }
#ifdef INPORT
	       else if(i>(beat*2 - port_sub)) {
		 overtones[j].si +=
		 (port_val*overtones[j].r +
		  (32768-port_val)*overtones[(j+1)%16].r)>>15;
		  port_val -= port_delta;
	       }
#endif
	       else
		 overtones[j].si += overtones[j].r;
	       
	       overtones[j].si &= 1023;
	    }
	    waveform[i]=d;
	 }
/*	    playbuffer16[i] = (bit16)((d*lpan*vol)>>11);*/
/*	    playbuffer16[i+1] = (bit16)((d*rpan*vol)>>11);*/
#ifdef RAD

	 wi %= 1024;
#endif
#endif /* RECALC_WAVE */	 
	 playbuffer16[i] = (bit16)((playbuffer16[i]*envelope)>>15);
	 playbuffer16[i+1] = (bit16)((playbuffer16[i+1]*envelope)>>15);
	 if(i<attack_sub) envelope+=(attack_delta<<1);
	 else if(i>release_sub) envelope-=(release_delta<<1);
      }
#ifdef CRAP
      /* SUSTAIN */
      for(; i<release_sub; i+=2) {
	 playbuffer16[i] = (bit16)((waveform[wi]*lpan*vol)>>11);
	 playbuffer16[i+1] = (bit16)((waveform[wi++]*rpan*vol)>>11);
	 if(!(wi%=1024)) {
	    d=0;
	    for(j=0; j<16; j++) {
	       d+= (sinwave[overtones[j].si] * overtones[j].vol)>>5;
	       overtones[j].si += overtones[j].r;
	       overtones[j].si &= 1023;
	    }
	    waveform[i] = d;
	 }
#ifdef RAD
	 wi %= 1024;
#endif
      }
      /* RELEASE */
      for(; i<(beat*2); i+=2) {
	 playbuffer16[i] = (bit16)((waveform[wi]*lpan*vol)>>11);
	 playbuffer16[i+1] = (bit16)((waveform[wi++]*rpan*vol)>>11);
	 if(!(wi%=1024)) {
	    d=0;
	    for(j=0; j<16; j++) {
	       d+= (sinwave[overtones[j].si] * overtones[j].vol)>>5;
	       overtones[j].si += overtones[j].r;
	       overtones[j].si &= 1023;
	    }
	    waveform[i] = d;
	 }
	 
#ifdef RAD
	 wi %= 1024;
#endif
	 playbuffer16[i] = (bit16)((playbuffer16[i]*envelope)>>15);
	 playbuffer16[i+1] = (bit16)((playbuffer16[i+1]*envelope)>>15);
	 envelope-=(release_delta<<1);
      }
#endif /* CRAP */
   }
   
   else {
      for(i=0; i<playlength; i+=2) {
	 playbuffer16[i]=0;
	 playbuffer16[i+1]=0;
      }
   }
   
   
   write(dev_dsp, playbuffer16, (size_t)(beat<<2));
      
   seqi++;
   seqi%=16;
   
/*
   XQueryPointer(disp, win, &proot, &pchild,
		 &rootx, &rooty, &winx, &winy, &keys_buttons);
   
   dx= winx-lx; dy= winy-ly;
      
   lx=winx; ly=winy;
*/   

}


main(int argc, char **argv)
{  
   int i;
   Pixmap omap,freshmap;
      struct itimerval itv;
   
   srandom(time(NULL));

   SetupDisplay();   

   open_dev_dsp();

   generate_sinwave();

   chaos_harmonics();
   
   chaos();
   
   width=640;
   height=480;

   OpenWindow();
/*   win=Root;*/
   
/*   SetupColormap();*/
   
   freshmap=XCreatePixmap(disp, win, width, height, depth);

   for(i=0; i<playlength; i++) {
      playbuffer16[i++]=playbuffer16[i]=(short)0;
   }
   write(dev_dsp, playbuffer16, (size_t)(playlength<<1));
   
#ifdef _USE_SIGNAL
   fprintf(stderr, "Using signals for timing...\n");

   signal(SIGALRM, handler_alarm);
   
   itv.it_interval.tv_usec = FRAMETIME/32;
   itv.it_interval.tv_sec = 0;
   itv.it_value.tv_usec = FRAMETIME/32;
   itv.it_value.tv_sec = 0;
   
   setitimer(ITIMER_REAL, &itv, (struct itimerval *)0);    
   
   
   for(;;);

#else
/*   fprintf(stderr, "Using loop for timing\n");*/
   
   for(;;) {
      tick();
/*      usleep(1000);*/
/*      usleep(FRAMETIME);*/
   }
#endif
   
#ifdef CRAP
     {
      XNextEvent(disp, &report);
      
      switch(report.type) {
       case ButtonPress:
	 break;
       case MotionNotify:
	 break;
       case ButtonRelease:
#ifdef PROF
	 exit(0);
#endif
	 break;
       default:
	 break;
      }
      
      
/*      XFlush(disp);*/
      
/*      usleep(1000000/30);*/
     }
#endif /* CRAP */
   
}
